'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var stream = require('readable-stream');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var stream__default = /*#__PURE__*/_interopDefaultLegacy(stream);

function asCallback (fn) {
  if (typeof fn !== 'function') {
    throw new TypeError('A callback must be a function')
  }
  fn[Symbol.for('callback')] = true;
  return fn
}

function isCallback (fn) {
  return typeof fn === 'function' && Symbol.for('callback') in fn
}

function once (callback) {
  let called = false;
  return function (err, res) {
    if (called) return
    called = true;
    callback(err, res);
  }
}

function makeConcurrentMethods (concurrency, _write, _final) {
  if (!Number.isInteger(concurrency) || concurrency < 2) {
    throw new TypeError('Invalid concurrency')
  }

  // Active jobs counter
  let jobs = 0;

  // First encountered error
  let error;

  // Require more chunks callback
  let next;

  // End stream callback
  let close;

  function concurrentWrite (chunk, encoding, callback) {
    if (!error) {
      jobStart(this, chunk, encoding);
    }

    if (!error && jobs < concurrency) {
      // Keep requiring more chunks
      callback();
    } else {
      // Save the current callback for jobs completion
      next = callback;
    }
  }

  function jobStart (stream, chunk, encoding) {
    jobs++;
    _write.call(
      stream,
      chunk,
      encoding,
      once((err, data) => jobEnd(stream, err, data))
    );
  }

  function jobEnd (stream, err, data) {
    jobs--;
    error = error || err;

    if (!error && data !== undefined && stream.push) {
      stream.push(data);
    }

    if (jobs <= 0) {
      const callback = close || next;
      if (callback) {
        callback(error);
      } else if (error) {
        stream.emit('error', error);
      }
    } else if (!error && next) {
      const callback = next;
      next = undefined;
      callback();
    }
  }

  function concurrentFinal (callback) {
    // Build close callback
    close = err => {
      if (err || !_final) {
        callback(err);
      } else {
        _final.call(this, callback);
      }
    };

    if (jobs <= 0) {
      close();
    }
  }

  return [concurrentWrite, concurrentFinal]
}

function patchWritable (writable, concurrency) {
  const [_write, _final] = makeConcurrentMethods(
    concurrency,
    writable._write,
    writable._final
  );
  writable._write = _write;
  writable._final = _final;
}

function patchTransform (transform, concurrency) {
  const [_transform, _flush] = makeConcurrentMethods(
    concurrency,
    transform._transform,
    transform._flush
  );
  transform._transform = _transform;
  transform._flush = _flush;
}

function handlePromise (fn) {
  if (typeof fn !== 'function') {
    return fn
  }
  return Object.defineProperty(
    function (...args) {
      const callback = args[args.length - 1];
      const value = fn.call(this, ...args);
      if (isPromiseLike(value)) {
        value.then(
          result => callback(null, result),
          callback
        );
      }
    },
    'name',
    { value: fn.name }
  )
}

function isPromiseLike (value) {
  return typeof Object(value).then === 'function'
}

function createReadMethod (asyncRead) {
  let reading = false;

  return function _read (size) {
    if (reading) {
      return
    }
    reading = true;

    const callback = once((err, data) => {
      reading = false;

      // Detects manual .push(null) and .destroy()
      const alive = !this._readableState.closed && !this._readableState.ended;
      if (!alive) {
        return
      }

      if (err) {
        // This will set ._readableState.closed to true
        return this.destroy(err)
      }

      if (data !== undefined) {
        // If null, this will set ._readableState.ended to true
        this.push(data);
      }

      if (
        data !== null &&
        this.readableFlowing &&
        this.readableLength < this.readableHighWaterMark
      ) {
        // We have room inside the buffer
        process.nextTick(() => this._read(size));
      }
    });

    const out = asyncRead.call(this, size, callback);
    if (isPromiseLike(out)) {
      out.then(
        result => callback(null, result),
        callback
      );
    }
  }
}

class Duplex extends stream__default["default"].Duplex {
  static from (...args) {
    return stream__default["default"].Duplex.from(...args)
  }

  static fromWeb (...args) {
    return stream__default["default"].Duplex.fromWeb(...args)
  }

  static toWeb (...args) {
    return stream__default["default"].Duplex.toWeb(...args)
  }

  constructor (options) {
    super(options);

    if (typeof Object(options).asyncRead === 'function') {
      this._asyncRead = options.asyncRead;
    }

    this._write = handlePromise(this._write);
    this._writev = handlePromise(this._writev);
    this._destroy = handlePromise(this._destroy);
    this._final = handlePromise(this._final);
    this._construct = handlePromise(this._construct);

    if (this._asyncRead) {
      this._read = createReadMethod(this._asyncRead);
    }

    const concurrency = Object(options).concurrency;
    if (Number.isInteger(concurrency) && concurrency >= 2) {
      patchWritable(this, concurrency);
    }
  }
}

function finished (...args) {
  return typeof args[args.length - 1] === 'function'
    ? stream__default["default"].finished(...args)
    : stream__default["default"].promises.finished(...args)
}

function isNodeStream (obj) {
  return isReadableStream(obj) || isWritableStream(obj)
}

function isReadableStream (value) {
  return value instanceof stream.Stream && typeof value.push === 'function'
}

function isWritableStream (value) {
  return value instanceof stream.Stream && typeof value.write === 'function'
}

function isDuplexStream (value) {
  return isReadableStream(value) && isWritableStream(value)
}

function pipeline (...args) {
  return isCallback(args[args.length - 1])
    ? stream__default["default"].pipeline(...args)
    : stream__default["default"].promises.pipeline(...args)
}

class Readable extends stream__default["default"].Readable {
  static from (...args) {
    return stream__default["default"].Readable.from(...args)
  }

  static fromWeb (...args) {
    return stream__default["default"].Readable.fromWeb(...args)
  }

  static isDisturbed (...args) {
    return stream__default["default"].Readable.isDisturbed(...args)
  }

  static toWeb (...args) {
    return stream__default["default"].Readable.toWeb(...args)
  }

  constructor (options) {
    super(options);

    if (typeof Object(options).asyncRead === 'function') {
      this._asyncRead = options.asyncRead;
    }

    this._destroy = handlePromise(this._destroy);
    this._construct = handlePromise(this._construct);

    if (this._asyncRead) {
      if (this._read !== Readable.prototype._read) {
        throw new Error('The _read method cannot be defined along with the _asyncRead method')
      }
      this._read = createReadMethod(this._asyncRead);
    }
  }
}

class Transform extends stream__default["default"].Transform {
  constructor (options) {
    super(options);

    this._destroy = handlePromise(this._destroy);
    this._construct = handlePromise(this._construct);
    this._transform = handlePromise(this._transform);
    this._flush = handlePromise(this._flush);

    const concurrency = Object(options).concurrency;
    if (Number.isInteger(concurrency) && concurrency >= 2) {
      patchTransform(this, concurrency);
    }
  }
}

class Writable extends stream__default["default"].Writable {
  static fromWeb (...args) {
    return stream__default["default"].Writable.fromWeb(...args)
  }

  static toWeb (...args) {
    return stream__default["default"].Writable.toWeb(...args)
  }

  constructor (options) {
    super(options);

    this._write = handlePromise(this._write);
    this._writev = handlePromise(this._writev);
    this._destroy = handlePromise(this._destroy);
    this._final = handlePromise(this._final);
    this._construct = handlePromise(this._construct);

    const concurrency = Object(options).concurrency;
    if (Number.isInteger(concurrency) && concurrency >= 2) {
      patchWritable(this, concurrency);
    }
  }
}

Object.defineProperty(exports, 'PassThrough', {
  enumerable: true,
  get: function () { return stream.PassThrough; }
});
Object.defineProperty(exports, 'Stream', {
  enumerable: true,
  get: function () { return stream.Stream; }
});
Object.defineProperty(exports, 'addAbortSignal', {
  enumerable: true,
  get: function () { return stream.addAbortSignal; }
});
Object.defineProperty(exports, 'compose', {
  enumerable: true,
  get: function () { return stream.compose; }
});
Object.defineProperty(exports, 'isErrored', {
  enumerable: true,
  get: function () { return stream.isErrored; }
});
Object.defineProperty(exports, 'isReadable', {
  enumerable: true,
  get: function () { return stream.isReadable; }
});
exports.Duplex = Duplex;
exports.Readable = Readable;
exports.Transform = Transform;
exports.Writable = Writable;
exports.asCallback = asCallback;
exports.finished = finished;
exports.isCallback = isCallback;
exports.isDuplexStream = isDuplexStream;
exports.isNodeStream = isNodeStream;
exports.isReadableStream = isReadableStream;
exports.isWritableStream = isWritableStream;
exports.pipeline = pipeline;
